# Copyright 2024 NVIDIA Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from legate.core import get_legate_runtime, types as ty

from .._utils import is_np2
from ..config import CuPyNumericOpCode
from ..runtime import runtime

if is_np2:
    from numpy.lib.array_utils import normalize_axis_index  # type: ignore
else:
    from numpy.core.multiarray import normalize_axis_index  # type: ignore

if TYPE_CHECKING:
    from .._thunk.deferred import DeferredArray


def sort_flattened(
    output: DeferredArray, input: DeferredArray, argsort: bool, stable: bool
) -> None:
    flattened = cast("DeferredArray", input.reshape((input.size,), order="C"))

    # run sort flattened -- return 1D solution
    sort_result = cast(
        "DeferredArray",
        runtime.create_empty_thunk(
            flattened.shape, dtype=output.base.type, inputs=(flattened,)
        ),
    )
    sort_deferred(sort_result, flattened, argsort, stable=stable)
    output.base = sort_result.base
    output.numpy_array = None


def sort_swapped(
    output: DeferredArray,
    input: DeferredArray,
    argsort: bool,
    sort_axis: int,
    stable: bool,
) -> None:
    sort_axis = normalize_axis_index(sort_axis, input.ndim)

    # swap axes
    swapped = input.swapaxes(sort_axis, input.ndim - 1)

    swapped_copy = cast(
        "DeferredArray",
        runtime.create_empty_thunk(
            swapped.shape, dtype=input.base.type, inputs=(input, swapped)
        ),
    )
    swapped_copy.copy(swapped, deep=True)

    # run sort on last axis
    if argsort is True:
        sort_result = cast(
            "DeferredArray",
            runtime.create_empty_thunk(
                swapped_copy.shape,
                dtype=output.base.type,
                inputs=(swapped_copy,),
            ),
        )
        sort_deferred(sort_result, swapped_copy, argsort, stable=stable)
        output.base = sort_result.swapaxes(input.ndim - 1, sort_axis).base
        output.numpy_array = None
    else:
        sort_deferred(swapped_copy, swapped_copy, argsort, stable=stable)
        output.base = swapped_copy.swapaxes(input.ndim - 1, sort_axis).base
        output.numpy_array = None


def sort_task(
    output: DeferredArray, input: DeferredArray, argsort: bool, stable: bool
) -> None:
    legate_runtime = get_legate_runtime()
    task = legate_runtime.create_auto_task(
        output.library, CuPyNumericOpCode.SORT
    )

    uses_unbound_output = runtime.num_procs > 1 and input.ndim == 1

    task.add_input(input.base)
    if uses_unbound_output:
        unbound = runtime.create_unbound_thunk(dtype=output.base.type, ndim=1)
        task.add_output(unbound.base)
    else:
        task.add_output(output.base)
        task.add_alignment(output.base, input.base)

    if runtime.num_gpus > 1:
        task.add_nccl_communicator()
    elif runtime.num_gpus == 0 and runtime.num_procs > 1:
        task.add_cpu_communicator()

    task.add_scalar_arg(argsort, ty.bool_)  # return indices flag
    task.add_scalar_arg(input.base.shape, (ty.int64,))
    task.add_scalar_arg(stable, ty.bool_)
    task.execute()

    if uses_unbound_output:
        output.base = unbound.base
        output.numpy_array = None


def sort_deferred(
    output: DeferredArray,
    input: DeferredArray,
    argsort: bool,
    axis: int | None = -1,
    stable: bool = False,
) -> None:
    if axis is None and input.ndim > 1:
        sort_flattened(output, input, argsort, stable)
    else:
        if axis is None:
            computed_axis = 0
        else:
            computed_axis = normalize_axis_index(axis, input.ndim)

        if computed_axis == input.ndim - 1:
            sort_task(output, input, argsort, stable)
        else:
            sort_swapped(output, input, argsort, computed_axis, stable)
